#include <cstdlib>
#include <iomanip>
#include <iostream>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <cuda_runtime_api.h>

#include "mmult.h"

int main()
{
    // .
    const int m = BLOCK_SIZE * 12, n = BLOCK_SIZE * 8, k = BLOCK_SIZE * 14;
    //   .
    const int repeats = 10;

    // a   m x n
    float * a = new float[m * n];
    float * a_gpu;
    cudaMalloc((void **) &a_gpu, m * n * sizeof(float));

    // b   n x k
    float * b = new float[n * k];
    float * b_gpu;
    cudaMalloc((void **) &b_gpu, n * k * sizeof(float));

    // c   m x k
    float * c = new float[m * k];
    float * c_verify = new float[m * k];
    float * c_gpu;
    cudaMalloc((void **) &c_gpu, m * k * sizeof(float));

    //    .
    for (int i = 0; i < m * n; ++i)
        a[i] = float(rand()) / RAND_MAX;
    for (int i = 0; i < n * k; ++i)
        b[i] = float(rand()) / RAND_MAX;

    //     GPU.
    cudaMemcpy(a_gpu, a, m * n * sizeof(float), cudaMemcpyHostToDevice);
    cudaMemcpy(b_gpu, b, n * k * sizeof(float), cudaMemcpyHostToDevice);

    //       .
    LARGE_INTEGER before, after, freq;
    QueryPerformanceFrequency(&freq);
    std::cout << std::fixed << std::setprecision(3);

    //  CPU-.
    QueryPerformanceCounter(&before);
    for (int i = 0; i < repeats; ++i)
        mmult(m, n, k, a, b, c);
    QueryPerformanceCounter(&after);
    double cpu_time = (after.QuadPart - before.QuadPart) / double(freq.QuadPart);
    std::cout << "The CPU version took " << cpu_time << " seconds.\n\n";

    //   GPU-.
    QueryPerformanceCounter(&before);
    for (int i = 0; i < repeats; ++i)
        mmult_gpu(m, n, k, a_gpu, b_gpu, c_gpu);
    // ,     .
    cudaThreadSynchronize();
    QueryPerformanceCounter(&after);
    double gpu_time = (after.QuadPart - before.QuadPart) / double(freq.QuadPart);
    std::cout << "The naive GPU version took " << gpu_time << " seconds.\n";
    std::cout << "Speedup: " << cpu_time / gpu_time << " times.\n";

    //        .
    cudaMemcpy(c_verify, c_gpu, m * k * sizeof *c_gpu,
        cudaMemcpyDeviceToHost);

    //     c  c_verify.
    float difference = 0;
    for (int i = 0; i < m * k; ++i)
        difference += (c[i] - c_verify[i]) * (c[i] - c_verify[i]);
    if (difference < 1e-5f)
        std::cout << "Test passed.\n";
    else
        std::cout << "Test failed (diff = " << difference << ").\n";
    std::cout << "\n";

    //     c_gpu .
    cudaMemset(c_gpu, 0, m * k * sizeof(float));

    //   GPU-.
    QueryPerformanceCounter(&before);
    for (int i = 0; i < repeats; ++i)
        mmult_gpu_opt(m, n, k, a_gpu, b_gpu, c_gpu);
    // ,     .
    cudaThreadSynchronize();
    QueryPerformanceCounter(&after);
    double opt_gpu_time = (after.QuadPart - before.QuadPart) / double(freq.QuadPart);
    std::cout << "The optimized GPU version took " << opt_gpu_time << " seconds.\n";
    std::cout << "Speedup: " << cpu_time / opt_gpu_time << " times.\n";

    //        .
    cudaMemcpy(c_verify, c_gpu, m * k * sizeof *c_gpu,
        cudaMemcpyDeviceToHost);

    //     c  c_verify.
    difference = 0;
    for (int i = 0; i < m * k; ++i)
        difference += (c[i] - c_verify[i]) * (c[i] - c_verify[i]);
    if (difference < 1e-5f)
        std::cout << "Test passed.\n";
    else
        std::cout << "Test failed (diff = " << difference << ").\n";

    //  .
    cudaFree(c_gpu);
    delete [] c_verify;
    delete [] c;
    cudaFree(b_gpu);
    delete [] b;
    cudaFree(a_gpu);
    delete [] a;
}
